//-----------------------------------------------------------------------------
//
//  npmapfile.c
//
//  ANTz application based on NPE is free software
//	NPE is the Neural Physics Engine that drives the interactions
//
//	ANTz hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
//	Copyright (c) 2010-2011, Shane Saxon
//	All rights reserved.
//	Redistribution and use in source and binary forms, with or without 
//	modification, are permitted provided that the following conditions are met:
//		Redistributions of source code must retain the above copyright notice, 
//		this list of conditions and the following disclaimer.
//		Redistributions in binary form must reproduce the above copyright 
//		notice, this list of conditions and the following disclaimer in the 
//		documentation and/or other materials provided with the distribution.
//		Neither the name of the Shane Saxon nor the names of its contributors 
//		may be used to endorse or promote products derived from this software 
//		without specific prior written permission.
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
//	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//	POSSIBILITY OF SUCH DAMAGE.
//
//------------------------------------------------------------------------------

#include "npmapfile.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "../npdata.h"
#include "../data/npnode.h"
#include "../os/npos.h"
#include "../io/npfile.h"


void npLoadChannelFile (char* buffer, int wordSize, int size, void* dataRef);


//-----------------------------------------------------------------------------
void npInitMapFile (void* dataRef)
{
	return;
}

//-----------------------------------------------------------------------------
void npCloseMapFile (void* dataRef)
{
	return;
}

//-----------------------------------------------------------------------------
// mode usage: 3 parameters "r/w/a" followed by optional "b" and optional "+"
// example mode = "rb+"
// r open file for reading
// w open file for overwriting. Create new file if doesn't already exist
// a open file for appending data. 
// b open file as a binary file
// + open file for read and writing
//-----------------------------------------------------------------------------
FILE* npOpenMapFile (const char* fileName, const char* mode, void* dataRef)
{
	FILE* file = NULL;

	file = fopen(fileName, mode); 

	if ( file == NULL )
	{
		perror("Failed to open/create file");
		return NULL;
	}

	// add code to add the file ptr to the data->map							debug, zz

	return file;
}

//-----------------------------------------------------------------------------
FILE* npNewMapFile (const char* fileName, const char* mode, void* dataRef)
{
	return NULL;
}


//-----------------------------------------------------------------------------
int npReadMapFile (void* readBuffer, int elementSize, int elementCount, 
				 FILE* file, void* dataRef)
{
	int bytes = 0;

	bytes = fread (readBuffer, elementSize, elementCount, file);
	
	if ( bytes == 0 )
		printf("Failed Reading\n");
	else
		printf("Success Reading\n");

	return bytes;
}


//-----------------------------------------------------------------------------
FILE* npDialogMapFile (const char* fileName, int dialogType, void* dataRef)
{	
	
	switch (dialogType)
	{
		case kNPfileDialogNew : ; break;

		case kNPfileDialogOpen : ; break;
		case kNPfileDialogClose : ; break;

		case kNPfileDialogSave : ; break;
		case kNPfileDialogSaveAs : ; break;

		case kNPfileDialogImport : ; break;
		case kNPfileDialogExport : ; break;

		default : break;
	}

	return NULL;
}


//-----------------------------------------------------------------------------
int npWriteMapFile (const void* str, int wordSize, int size, FILE* file, void* dataRef)
{
	return fwrite( str, wordSize, strlen(str), file );
}



void npSaveMapFile (FILE* file, void* dataRef)
{
	return;
}


//-----------------------------------------------------------------------------
int npImportMapFile (FILE* file, int type, void* dataRef)
{
	int size = 0;
	return size;
}

//------------------------------------------------------------------------------
int npMapTraverseTree (char* writeBuffer, NPnodePtr node, void* dataRef)
{
	int i = 0;
	int count = 0;

	// if not root then write the data, root has already been written
	if (node->branchLevel > 0)
			count += npWriteNode (writeBuffer, node, dataRef);


	//recursively calls this function for all children
	for (i=0; i < node->childCount; i++)//node->childCount; j++)
	{
		if (node->child[i] != NULL)
			count += npMapTraverseTree ((writeBuffer+count), node->child[i], dataRef);	
	}

	return count;
}

//-----------------------------------------------------------------------------
// write the header structure
int npWriteMapHeader (char* bufferIndex, int count, int type, void* dataRef)
{
	int i = 0;
	int itemCount = 0;

	NPmapTypePtr mapPtr= NULL;

	pData data = (pData) dataRef;

	count = 0;


	// need first one without comma
	count += sprintf ((bufferIndex + count), "%s", data->map.typeMapNode[0].name);

	switch (type)
	{
		case kNPnode: itemCount = 46; break;
		case kNPcamera: itemCount = 18; break;
		case kNPgrid: itemCount = 6; break;
		case kNPpin: itemCount = 13; break;

		default : break;
	}

	for (i=1; i < itemCount ;i++)
	{
		// insert a comma for CSV and JSON files
		count += sprintf ((bufferIndex + count), ",");

		switch (type)
		{
			case kNPnode:
		

		switch (data->map.typeMapNode[i].type)
		{
			case 0 :				// same as default
				count += sprintf ((bufferIndex + count), "%s", 
									data->map.typeMapNode[i].name );
				break;
			case kNPfloatXYZ : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );
				break;
			case kNPintXYZ : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );
				break;
			case kNPboolXYZ : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );	
				break;
			case kNPfloatXYZA : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z,%s a", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );
			case kNPfloatXYZS : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z,%s s", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );	
				break;
			case kNPboolXYZS : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z,%s s", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );	
				break;
			case kNPubyteRGBA : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b,%s a", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );	
				break;
			case kNPubyteRGB : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );	
				break;
			case kNPfloatRGBA : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b,%s a", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );	
				break;
			case kNPfloatRGB : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b", 
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name,
									data->map.typeMapNode[i].name );	
				break;

			default :				// same as case 0
				count += sprintf ((bufferIndex + count), "%s", 
									data->map.typeMapNode[i].name );
				break;
		}
		break;
			
		case kNPcamera:
			count += sprintf ((bufferIndex + count), "%s",
				data->map.typeMapCamera[i].name );
			break;
		case kNPgrid :
			switch (data->map.typeMapGrid[i].type)
			{
				case kNPfloatXYZ : 
				case kNPintXYZ : 
					count += sprintf ((bufferIndex + count), "%s x,%s y,%s z", 
									data->map.typeMapGrid[i].name,
									data->map.typeMapGrid[i].name,
									data->map.typeMapGrid[i].name );
					break;
				default :				// same as case 0
					count += sprintf ((bufferIndex + count), "%s", 
									data->map.typeMapGrid[i].name );
				break;
			}
			break;
		case kNPpin: 
			switch (data->map.typeMapPin[i].type)
			{
				case 0 :				// same as default
					count += sprintf ((bufferIndex + count), "%s", 
									data->map.typeMapPin[i].name );
					break;
				case kNPfloatXYZ : 
					count += sprintf ((bufferIndex + count), "%s x,%s y,%s z", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );
					break;
				case kNPintXYZ : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );
				break;
			case kNPboolXYZ : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );	
				break;
			case kNPfloatXYZA : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z,%s a", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );
			case kNPfloatXYZS : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z,%s s", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );	
				break;
			case kNPboolXYZS : 
				count += sprintf ((bufferIndex + count), "%s x,%s y,%s z,%s s", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );	
				break;
			case kNPubyteRGBA : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b,%s a", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );	
				break;
			case kNPubyteRGB : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );	
				break;
			case kNPfloatRGBA : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b,%s a", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );	
				break;
			case kNPfloatRGB : 
				count += sprintf ((bufferIndex + count), "%s r,%s g,%s b", 
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name,
									data->map.typeMapPin[i].name );	
				break;

			default :				// same as case 0
				count += sprintf ((bufferIndex + count), "%s", 
									data->map.typeMapPin[i].name );
				break;
		}
		
		break;

		default : break;
		}


	}

	count += sprintf ((bufferIndex + count), "\n");

	return count;
}

//-----------------------------------------------------------------------------
int npGetMapFile (char* buffer, int count, void* nodeRef, void* dataRef)
{
	int i = 0;
	int type = 0;
	pData data = (pData) dataRef;
	NPnodePtr node = (NPnodePtr) nodeRef;


	data->io.write[kNPnode] = buffer;
	data->io.write[kNPcamera] = buffer + ((kNPmapFileWriteBufferSize * 8) / 16);
	data->io.write[kNPgrid] = buffer + ((kNPmapFileWriteBufferSize * 9) / 16);
	data->io.write[kNPpin] = buffer + ((kNPmapFileWriteBufferSize * 10) / 16);
	data->io.write[kNPchild] = buffer + ((kNPmapFileWriteBufferSize * 12) / 16);
	data->io.write[kNPchannel] = buffer + ((kNPmapFileWriteBufferSize * 14) / 16);


//	for (i=1; i < 20 ;i++)		
//		count += sprintf ((bufferIndex + count), ",%s", data->map.typeMapNode[i].name);

	count += npWriteMapHeader ((buffer + count), count, kNPnode, dataRef);
	data->io.write[kNPnode] += count;
	
	count += npWriteMapHeader (data->io.write[kNPcamera], count, kNPcamera, dataRef);
	data->io.write[kNPcamera] += count;

	count += npWriteMapHeader (data->io.write[kNPgrid], count, kNPgrid, dataRef);
	data->io.write[kNPgrid] += count;

	count += npWriteMapHeader (data->io.write[kNPpin], count, kNPpin, dataRef);
	data->io.write[kNPpin] += count;

	count += npWriteMapHeader (data->io.write[kNPchild], count, kNPchild, dataRef);
	data->io.write[kNPchild] += count;

	count += npWriteMapHeader (data->io.write[kNPchannel], count, kNPchannel, dataRef);
	data->io.write[kNPchannel] += count;


	// iterate through data map and write out each node
	for (i=0; i < data->map.nodeRootCount; i++)
	{
		node = data->map.node[i];

		count += npWriteNode ((buffer + count), node, dataRef);

		if (node->childCount)
			count += npMapTraverseTree ((buffer + count), node, dataRef);
	}

	return count;
}

// iterate through nodes update translation rotation and scale
//-----------------------------------------------------------------------------
int npExportMapFile (FILE* file, int type, void* dataRef)
{
	int size = 0;

	return size;
}

//MB//

//-----------------------------------------------------------------------------
int npSeekToNextLine( char* buffer )	//Todo: Check Max
{
	char ch = 0;
	int curs = 0;

	while( ch != '\n' )		//Carraige Return at end of Line
	{
		ch = buffer[curs];
		curs++;
	}

	return curs;
}

//-----------------------------------------------------------------------------
//#define FIRSTCHAR 1141
#define MAXDEPTH 1024
void npLoadMapFile (char* buffer, int wordSize, int size, void* dataRef)
{
	int i = 0;
	int curs = 0;				//Cursor position in Buffer parsing source
	int scanNumRet = 0;			//sscanf return value. Number of successfuly scanned elements
	int numExpected = 0;		//Expected number for scanNumRet returned from sscanf
	int parentBranchLevel = 0;
	
	//Group #1
	int id, type, selected,	parentID, branchLevel, childIndex, childCount, 
		channel_x, channel_y, channel_z, channelIndex_x, channelIndex_y, channelIndex_z, 
		averageType_x, averageType_y, averageType_z, sampleInterval_x, sampleInterval_y, sampleInterval_z;
	//Group #2
	float rotate_angle, rotate_x, rotate_y, rotate_z, scale_x, scale_y, scale_z, translate_x, translate_y, translate_z,
		nodeOrigin_x, nodeOrigin_y, nodeOrigin_z, rotateRate_x, rotateRate_y, rotateRate_z, rotateRad_x, rotateRad_y, rotateRad_z;
	//Group #3
	float scaleRate_x, scaleRate_y, scaleRate_z, translateRate_x, translateRate_y, translateRate_z, 
		translateVect_x, translateVect_y, translateVect_z;
	//Group #4
	int surfaceType, shaderType;									
	float lineWidth, pointSize, circle;								
	int colorIndex, color_r, color_g, color_b, color_a, colorFade;
	//Group #5
	int hide, freeze, center, centerPerGrid, autoZoom_x, autoZoom_y, autoZoom_z, scroll, 
		triggerOnHi_x, triggerOnHi_y, triggerOnHi_z, triggerOnLow_x, triggerOnLow_y, triggerOnLow_z;
	//Group #6
	float triggerLevelHi_x, triggerLevelHi_y, triggerLevelHi_z, triggerLevelLow_x, triggerLevelLow_y, triggerLevelLow_z;
	int recordID;//int triggerCommand;		//zz debug
	float proximity_x, proximity_y, proximity_z, proximityRadius_x, proximityRadius_y, proximityRadius_z;
	//Group #7
	int drawCurrentValue_x, drawCurrentValue_y, drawCurrentValue_z, selectionSet_x, selectionSet_y, selectionSet_z, lineSize;


	//
	NPnodePtr node = NULL;
	NPnodePtr nodeChild = NULL;
	NPnodePtr nodeChildLevel[MAXDEPTH];
	pData data = (pData) dataRef;

	for( i = 0; i < MAXDEPTH; i++ )
		nodeChildLevel[i] = NULL;

//	printf( "%s\n", __FUNCTION__ );
//	printf( "size: %d\n", size ); 
//	printf( "buffer: %s\n", buffer ); 

//	buffer = &buffer[FIRSTCHAR];		//Seek to First "Real" char. After all the Text Labels. "1" is the first number. It's an ID
	//Seek through first two lines, delimited by carraige returns. 
	//File Must have exactly two lines ('n') before first node: (id: 1) 
	
	
	curs = npSeekToNextLine( buffer );	buffer = &buffer[curs];		//First  Line not used
	
	for( i = 0; i < 128; i++ )
					printf( "%c", buffer[i]);

		printf( "\n 2nd group \n", buffer[0] );

	curs = npSeekToNextLine( buffer );	buffer = &buffer[curs];		//Second Line not used
	
	for( i = 0; i < 128; i++ )
					printf( "%c", buffer[i]);

	printf( "First Real char:%c\n", buffer[0] );

	numExpected = 94;		//21 ints + 19 floats + 9 floats + 11 int+floats + 14 ints + 13 floats+int + 7 ints
	while( scanNumRet != -1 )
	{
		curs = 0;

		//Scan Whole Line in 
		scanNumRet = sscanf( buffer, 
			"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%d,%d,%f,%f,%f,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%f,%f,%f,%f,%f,%f,%d,%f,%f,%f,%f,%f,%f,%d,%d,%d,%d,%d,%d,%d", 
			&id, &type, &id, &selected, &parentID, &branchLevel, &id, &childIndex, &childCount, 
			&channel_x, &channel_y, &channel_z, &channelIndex_x, &channelIndex_y, &channelIndex_z,
			&averageType_x, &averageType_y, &averageType_z, &sampleInterval_x, &sampleInterval_y, &sampleInterval_z,
			&rotate_angle, &rotate_x, &rotate_y, &rotate_z, &scale_x, &scale_y, &scale_z, &translate_x, &translate_y, &translate_z,
			&nodeOrigin_x, &nodeOrigin_y, &nodeOrigin_z, &rotateRate_x, &rotateRate_y, &rotateRate_z, &rotateRad_x, &rotateRad_y, &rotateRad_z,
			&scaleRate_x, &scaleRate_y, &scaleRate_z, &translateRate_x, &translateRate_y, &translateRate_z, 
			&translateVect_x, &translateVect_y, &translateVect_z, 
			&surfaceType, &shaderType, &lineWidth, &pointSize, &circle, &colorIndex, &color_r, &color_g, &color_b, &color_a, &colorFade,
			&hide, &freeze, &center, &centerPerGrid, &autoZoom_x, &autoZoom_y, &autoZoom_z, &scroll, 
			&triggerOnHi_x, &triggerOnHi_y, &triggerOnHi_z, &triggerOnLow_x, &triggerOnLow_y, &triggerOnLow_z,
			&triggerLevelHi_x, &triggerLevelHi_y, &triggerLevelHi_z, &triggerLevelLow_x, &triggerLevelLow_y, &triggerLevelLow_z, &recordID,
			&proximity_x, &proximity_y, &proximity_z, &proximityRadius_x, &proximityRadius_y, &proximityRadius_z,
			&drawCurrentValue_x, &drawCurrentValue_y, &drawCurrentValue_z, &selectionSet_x, &selectionSet_y, &selectionSet_z, &lineSize
		);
		printf( " scanNumRet: %d   type: %d\n", scanNumRet, type );
		if( scanNumRet != numExpected )
		{
			if( scanNumRet == -1 )
			{
				printf( "End of File\n" );
				continue;
			}
			else
			{
				printf( "WARNING: Bad Format, scanNumRet: %d\n",scanNumRet );
	//			for( i = 0; i < MAXDEPTH; i++ )
	//				printf( "%c", buffer[i-1024]);
			}
		}
		if( lineSize != 420 )								// need to be flexible, debug, zz
		{
			printf( "WARNING: Unexpected lineSize: %d\n", lineSize );
		}

/*		printf( "id: %d\n", id );
		printf( "type: %d\n", type );
		printf( "branchLevel: %d\n", branchLevel );
		printf( "lineSize: %d\n", lineSize );
*/
		//Find Next Line
		curs = npSeekToNextLine( buffer );
		buffer = &buffer[curs];

		if( type == 8)				//workaround for 7-11 to 7-21 CSV files, debug zz
			type = kNodePin;

		//New Node
		if( type == kNodePin && branchLevel == 0 )		//Pin at Root Level
		{
//			printf( "ROOT PIN! id:%d\n", id );
			node = npNodeNew( kNodePin, NULL, dataRef );		//Create Pin

			node->type			= type;

//			node->data			= NULL;				// handled during node creation

			node->selected		= selected;

//			node->parent		= NULL;				// handled during node creation
//			node->branchLevel	= 0;				// handled during node creation

			node->childIndex	= childIndex;
//			node->childCount	= childCount;		// handled during node creation

			node->channel.x		= channel_x;
			node->channel.y		= channel_y;
			node->channel.z		= channel_z;
			
			node->channelIndex.x = channelIndex_x;
			node->channelIndex.y = channelIndex_y;
			node->channelIndex.z = channelIndex_z;

			node->averageType.x	= averageType_x;
			node->averageType.y	= averageType_y;
			node->averageType.z	= averageType_z;
			
			node->sampleInterval.x = sampleInterval_x;
			node->sampleInterval.y = sampleInterval_y;
			node->sampleInterval.z = sampleInterval_z;

			node->rotate.angle	= rotate_angle;
			node->rotate.x		= rotate_x;
			node->rotate.y		= rotate_y;
			node->rotate.z		= rotate_z;
			
			node->scale.x		= scale_x;
			node->scale.y		= scale_y;
			node->scale.z		= scale_z;
			
			node->translate.x	= translate_x;
			node->translate.y	= translate_y;
			node->translate.z	= translate_z;

			node->nodeOrigin.x	= nodeOrigin_x;
			node->nodeOrigin.y	= nodeOrigin_y;
			node->nodeOrigin.z	= nodeOrigin_z;
			
			node->rotateRate.x	= rotateRate_x;
			node->rotateRate.y	= rotateRate_y;
			node->rotateRate.z	= rotateRate_z;

			node->rotateRad.x	= rotateRad_x;
			node->rotateRad.y	= rotateRad_y;
			node->rotateRad.z	= rotateRad_z;
			
			node->scaleRate.x	= scaleRate_x;
			node->scaleRate.y	= scaleRate_y;
			node->scaleRate.z	= scaleRate_z;

			node->translateRate.x = translateRate_x;
			node->translateRate.y = translateRate_y;
			node->translateRate.z = translateRate_z;

			node->translateVect.x = translateVect_x;
			node->translateVect.y = translateVect_y;
			node->translateVect.z = translateVect_z;
			
			node->surfaceType	= surfaceType;
			node->shaderType	= shaderType;

			node->lineWidth		= lineWidth;
			node->pointSize		= pointSize;
			node->circle		= circle;

			node->colorIndex	= colorIndex;
			
			node->color.r		= color_r;
			node->color.g		= color_g;
			node->color.b		= color_b;
			node->color.a		= color_a;

			node->colorFade		= colorFade;

			node->hide			= hide;
			node->freeze		= freeze;
			node->center		= center;
			node->centerPerGrid	= centerPerGrid;

			node->autoZoom.x	= autoZoom_x;
			node->autoZoom.y	= autoZoom_y;
			node->autoZoom.z	= autoZoom_z;

			node->scroll		= scroll;
			
			node->triggerOnHi.x	= triggerOnHi_x;
			node->triggerOnHi.y	= triggerOnHi_y;
			node->triggerOnHi.z	= triggerOnHi_z;

			node->triggerOnLow.x = triggerOnLow_x;
			node->triggerOnLow.y = triggerOnLow_y;
			node->triggerOnLow.z = triggerOnLow_z;
			
			node->triggerLevelHi.x = triggerLevelHi_x;
			node->triggerLevelHi.y = triggerLevelHi_y;
			node->triggerLevelHi.z = triggerLevelHi_z;

			node->triggerLevelLow.x = triggerLevelLow_x;
			node->triggerLevelLow.y = triggerLevelLow_y;
			node->triggerLevelLow.z = triggerLevelLow_z;

			//node->triggerCommand = triggerCommand;		//zz debug
			node->recordID = (int)recordID;
			
			node->proximity.x	= (int)proximity_x;
			node->proximity.y	= (int)proximity_y;
			node->proximity.z	= (int)proximity_z;
			
			node->proximityRadius.x = proximityRadius_x;
			node->proximityRadius.y = proximityRadius_y;
			node->proximityRadius.z = proximityRadius_z;

			node->drawCurrentValue.x = drawCurrentValue_x;
			node->drawCurrentValue.y = drawCurrentValue_y;
			node->drawCurrentValue.z = drawCurrentValue_z;

			node->selectionSet.x	= selectionSet_x;
			node->selectionSet.y	= selectionSet_y;
			node->selectionSet.z	= selectionSet_z;

			node->size = lineSize;
		}
	
		//New Primary Child Toroid
		if( type == kNodePin && branchLevel == 1 )		//First Level Toroid 
		{
//			printf( "Child PIN! id:%d\n", id );
			nodeChild = node->child[0];		//created same time as root, debug, zz

			nodeChild->type			= type;

//			node->data				= NULL;				// handled during node creation

			nodeChild->selected		= selected;

//			node->parent			= NULL;				// handled during node creation
//			node->branchLevel		= 0;				// handled during node creation

			nodeChild->childIndex	= childIndex;
//			node->childCount		= childCount;		// handled during node creation

			nodeChild->channel.x		= channel_x;
			nodeChild->channel.y		= channel_y;
			nodeChild->channel.z		= channel_z;
			
			nodeChild->channelIndex.x	= channelIndex_x;
			nodeChild->channelIndex.y	= channelIndex_y;
			nodeChild->channelIndex.z	= channelIndex_z;

			nodeChild->averageType.x	= averageType_x;
			nodeChild->averageType.y	= averageType_y;
			nodeChild->averageType.z	= averageType_z;
			
			nodeChild->sampleInterval.x = sampleInterval_x;
			nodeChild->sampleInterval.y = sampleInterval_y;
			nodeChild->sampleInterval.z = sampleInterval_z;

			nodeChild->rotate.angle	= rotate_angle;
			nodeChild->rotate.x		= rotate_x;
			nodeChild->rotate.y		= rotate_y;
			nodeChild->rotate.z		= rotate_z;
			
			nodeChild->scale.x		= scale_x;
			nodeChild->scale.y		= scale_y;
			nodeChild->scale.z		= scale_z;
			
			nodeChild->translate.x	= translate_x;
			nodeChild->translate.y	= translate_y;
			nodeChild->translate.z	= translate_z;

			nodeChild->nodeOrigin.x	= nodeOrigin_x;
			nodeChild->nodeOrigin.y	= nodeOrigin_y;
			nodeChild->nodeOrigin.z	= nodeOrigin_z;
			
			nodeChild->rotateRate.x	= rotateRate_x;
			nodeChild->rotateRate.y	= rotateRate_y;
			nodeChild->rotateRate.z	= rotateRate_z;

			nodeChild->rotateRad.x	= rotateRad_x;
			nodeChild->rotateRad.y	= rotateRad_y;
			nodeChild->rotateRad.z	= rotateRad_z;
			
			nodeChild->scaleRate.x	= scaleRate_x;
			nodeChild->scaleRate.y	= scaleRate_y;
			nodeChild->scaleRate.z	= scaleRate_z;

			nodeChild->translateRate.x = translateRate_x;
			nodeChild->translateRate.y = translateRate_y;
			nodeChild->translateRate.z = translateRate_z;

			nodeChild->translateVect.x = translateVect_x;
			nodeChild->translateVect.y = translateVect_y;
			nodeChild->translateVect.z = translateVect_z;
			
			nodeChild->surfaceType	= surfaceType;
			nodeChild->shaderType	= shaderType;

			nodeChild->lineWidth	= lineWidth;
			nodeChild->pointSize	= pointSize;
			nodeChild->circle		= circle;

			nodeChild->colorIndex	= colorIndex;
			
			nodeChild->color.r		= color_r;
			nodeChild->color.g		= color_g;
			nodeChild->color.b		= color_b;
			nodeChild->color.a		= color_a;

			nodeChild->colorFade	= colorFade;

			nodeChild->hide			= hide;
			nodeChild->freeze		= freeze;
			nodeChild->center		= center;
			nodeChild->centerPerGrid	= centerPerGrid;

			nodeChild->autoZoom.x	= autoZoom_x;
			nodeChild->autoZoom.y	= autoZoom_y;
			nodeChild->autoZoom.z	= autoZoom_z;

			nodeChild->scroll		= scroll;
			
			nodeChild->triggerOnHi.x	= triggerOnHi_x;
			nodeChild->triggerOnHi.y	= triggerOnHi_y;
			nodeChild->triggerOnHi.z	= triggerOnHi_z;

			nodeChild->triggerOnLow.x	= triggerOnLow_x;
			nodeChild->triggerOnLow.y	= triggerOnLow_y;
			nodeChild->triggerOnLow.z	= triggerOnLow_z;
			
			nodeChild->triggerLevelHi.x = triggerLevelHi_x;
			nodeChild->triggerLevelHi.y = triggerLevelHi_y;
			nodeChild->triggerLevelHi.z = triggerLevelHi_z;

			nodeChild->triggerLevelLow.x = triggerLevelLow_x;
			nodeChild->triggerLevelLow.y = triggerLevelLow_y;
			nodeChild->triggerLevelLow.z = triggerLevelLow_z;

			//node->triggerCommand = triggerCommand;		//zz debug
			nodeChild->recordID = (int)recordID;
			
			nodeChild->proximity.x = (int)proximity_x;
			nodeChild->proximity.y = (int)proximity_y;
			nodeChild->proximity.z = (int)proximity_z;
			
			nodeChild->proximityRadius.x = proximityRadius_x;
			nodeChild->proximityRadius.y = proximityRadius_y;
			nodeChild->proximityRadius.z = proximityRadius_z;

			nodeChild->drawCurrentValue.x = drawCurrentValue_x;
			nodeChild->drawCurrentValue.y = drawCurrentValue_y;
			nodeChild->drawCurrentValue.z = drawCurrentValue_z;

			nodeChild->selectionSet.x = selectionSet_x;
			nodeChild->selectionSet.y = selectionSet_y;
			nodeChild->selectionSet.z = selectionSet_z;

			nodeChild->size = lineSize;
		}

		//New Child Toroid
		if( type == kNodePin && branchLevel == 2 )		//2nd Level Toroid 
		{
//			printf( "Child PIN! id:%d\n", id );
			nodeChild = npNodeNew (kNodePin, node->child[0], dataRef );
			nodeChildLevel[branchLevel] = nodeChild;	//compatible with traverse tree

			nodeChild->type			= type;

//			node->data				= NULL;				// handled during node creation

			nodeChild->selected		= selected;

//			node->parent			= NULL;				// handled during node creation
//			node->branchLevel		= 0;				// handled during node creation

			nodeChild->childIndex	= childIndex;
//			node->childCount		= childCount;		// handled during node creation

			nodeChild->channel.x		= channel_x;
			nodeChild->channel.y		= channel_y;
			nodeChild->channel.z		= channel_z;
			
			nodeChild->channelIndex.x	= channelIndex_x;
			nodeChild->channelIndex.y	= channelIndex_y;
			nodeChild->channelIndex.z	= channelIndex_z;

			nodeChild->averageType.x	= averageType_x;
			nodeChild->averageType.y	= averageType_y;
			nodeChild->averageType.z	= averageType_z;
			
			nodeChild->sampleInterval.x = sampleInterval_x;
			nodeChild->sampleInterval.y = sampleInterval_y;
			nodeChild->sampleInterval.z = sampleInterval_z;

			nodeChild->rotate.angle	= rotate_angle;
			nodeChild->rotate.x		= rotate_x;
			nodeChild->rotate.y		= rotate_y;
			nodeChild->rotate.z		= rotate_z;
			
			nodeChild->scale.x		= scale_x;
			nodeChild->scale.y		= scale_y;
			nodeChild->scale.z		= scale_z;
			
			nodeChild->translate.x	= translate_x;
			nodeChild->translate.y	= translate_y;
			nodeChild->translate.z	= translate_z;

			nodeChild->nodeOrigin.x	= nodeOrigin_x;
			nodeChild->nodeOrigin.y	= nodeOrigin_y;
			nodeChild->nodeOrigin.z	= nodeOrigin_z;
			
			nodeChild->rotateRate.x	= rotateRate_x;
			nodeChild->rotateRate.y	= rotateRate_y;
			nodeChild->rotateRate.z	= rotateRate_z;

			nodeChild->rotateRad.x	= rotateRad_x;
			nodeChild->rotateRad.y	= rotateRad_y;
			nodeChild->rotateRad.z	= rotateRad_z;
			
			nodeChild->scaleRate.x	= scaleRate_x;
			nodeChild->scaleRate.y	= scaleRate_y;
			nodeChild->scaleRate.z	= scaleRate_z;

			nodeChild->translateRate.x = translateRate_x;
			nodeChild->translateRate.y = translateRate_y;
			nodeChild->translateRate.z = translateRate_z;

			nodeChild->translateVect.x = translateVect_x;
			nodeChild->translateVect.y = translateVect_y;
			nodeChild->translateVect.z = translateVect_z;
			
			nodeChild->surfaceType	= surfaceType;
			nodeChild->shaderType	= shaderType;

			nodeChild->lineWidth	= lineWidth;
			nodeChild->pointSize	= pointSize;
			nodeChild->circle		= circle;

			nodeChild->colorIndex	= colorIndex;
			
			nodeChild->color.r		= color_r;
			nodeChild->color.g		= color_g;
			nodeChild->color.b		= color_b;
			nodeChild->color.a		= color_a;

			nodeChild->colorFade	= colorFade;

			nodeChild->hide			= hide;
			nodeChild->freeze		= freeze;
			nodeChild->center		= center;
			nodeChild->centerPerGrid	= centerPerGrid;

			nodeChild->autoZoom.x	= autoZoom_x;
			nodeChild->autoZoom.y	= autoZoom_y;
			nodeChild->autoZoom.z	= autoZoom_z;

			nodeChild->scroll		= scroll;
			
			nodeChild->triggerOnHi.x	= triggerOnHi_x;
			nodeChild->triggerOnHi.y	= triggerOnHi_y;
			nodeChild->triggerOnHi.z	= triggerOnHi_z;

			nodeChild->triggerOnLow.x	= triggerOnLow_x;
			nodeChild->triggerOnLow.y	= triggerOnLow_y;
			nodeChild->triggerOnLow.z	= triggerOnLow_z;
			
			nodeChild->triggerLevelHi.x = triggerLevelHi_x;
			nodeChild->triggerLevelHi.y = triggerLevelHi_y;
			nodeChild->triggerLevelHi.z = triggerLevelHi_z;

			nodeChild->triggerLevelLow.x = triggerLevelLow_x;
			nodeChild->triggerLevelLow.y = triggerLevelLow_y;
			nodeChild->triggerLevelLow.z = triggerLevelLow_z;

			//node->triggerCommand = triggerCommand;		//zz debug
			nodeChild->recordID = (int)recordID;
			
			nodeChild->proximity.x = (int)proximity_x;
			nodeChild->proximity.y = (int)proximity_y;
			nodeChild->proximity.z = (int)proximity_z;
			
			nodeChild->proximityRadius.x = proximityRadius_x;
			nodeChild->proximityRadius.y = proximityRadius_y;
			nodeChild->proximityRadius.z = proximityRadius_z;

			nodeChild->drawCurrentValue.x = drawCurrentValue_x;
			nodeChild->drawCurrentValue.y = drawCurrentValue_y;
			nodeChild->drawCurrentValue.z = drawCurrentValue_z;

			nodeChild->selectionSet.x = selectionSet_x;
			nodeChild->selectionSet.y = selectionSet_y;
			nodeChild->selectionSet.z = selectionSet_z;

			nodeChild->size = lineSize;
		}

		//New Sub Child Toroid
		if( type == kNodePin && branchLevel >= 3 )		//Second Level (or greater) Toroid Sub Child  
		{
			parentBranchLevel = branchLevel - 1;

//			printf( "Grand Child PIN! id:%d\n", id );
			nodeChild = npNodeNew (kNodePin, nodeChildLevel[parentBranchLevel], dataRef);
			nodeChildLevel[branchLevel] = nodeChild;

			nodeChild->type			= type;

//			node->data				= NULL;				// handled during node creation

			nodeChild->selected		= selected;

//			node->parent			= NULL;				// handled during node creation
//			node->branchLevel		= 0;				// handled during node creation

			nodeChild->childIndex	= childIndex;
//			node->childCount		= childCount;		// handled during node creation

			nodeChild->channel.x		= channel_x;
			nodeChild->channel.y		= channel_y;
			nodeChild->channel.z		= channel_z;
			
			nodeChild->channelIndex.x	= channelIndex_x;
			nodeChild->channelIndex.y	= channelIndex_y;
			nodeChild->channelIndex.z	= channelIndex_z;

			nodeChild->averageType.x	= averageType_x;
			nodeChild->averageType.y	= averageType_y;
			nodeChild->averageType.z	= averageType_z;
			
			nodeChild->sampleInterval.x = sampleInterval_x;
			nodeChild->sampleInterval.y = sampleInterval_y;
			nodeChild->sampleInterval.z = sampleInterval_z;

			nodeChild->rotate.angle	= rotate_angle;
			nodeChild->rotate.x		= rotate_x;
			nodeChild->rotate.y		= rotate_y;
			nodeChild->rotate.z		= rotate_z;
			
			nodeChild->scale.x		= scale_x;
			nodeChild->scale.y		= scale_y;
			nodeChild->scale.z		= scale_z;
			
			nodeChild->translate.x	= translate_x;
			nodeChild->translate.y	= translate_y;
			nodeChild->translate.z	= translate_z;

			nodeChild->nodeOrigin.x	= nodeOrigin_x;
			nodeChild->nodeOrigin.y	= nodeOrigin_y;
			nodeChild->nodeOrigin.z	= nodeOrigin_z;
			
			nodeChild->rotateRate.x	= rotateRate_x;
			nodeChild->rotateRate.y	= rotateRate_y;
			nodeChild->rotateRate.z	= rotateRate_z;

			nodeChild->rotateRad.x	= rotateRad_x;
			nodeChild->rotateRad.y	= rotateRad_y;
			nodeChild->rotateRad.z	= rotateRad_z;
			
			nodeChild->scaleRate.x	= scaleRate_x;
			nodeChild->scaleRate.y	= scaleRate_y;
			nodeChild->scaleRate.z	= scaleRate_z;

			nodeChild->translateRate.x = translateRate_x;
			nodeChild->translateRate.y = translateRate_y;
			nodeChild->translateRate.z = translateRate_z;

			nodeChild->translateVect.x = translateVect_x;
			nodeChild->translateVect.y = translateVect_y;
			nodeChild->translateVect.z = translateVect_z;
			
			nodeChild->surfaceType	= surfaceType;
			nodeChild->shaderType	= shaderType;

			nodeChild->lineWidth	= lineWidth;
			nodeChild->pointSize	= pointSize;
			nodeChild->circle		= circle;

			nodeChild->colorIndex	= colorIndex;
			
			nodeChild->color.r		= color_r;
			nodeChild->color.g		= color_g;
			nodeChild->color.b		= color_b;
			nodeChild->color.a		= color_a;

			nodeChild->colorFade	= colorFade;

			nodeChild->hide			= hide;
			nodeChild->freeze		= freeze;
			nodeChild->center		= center;
			nodeChild->centerPerGrid	= centerPerGrid;

			nodeChild->autoZoom.x	= autoZoom_x;
			nodeChild->autoZoom.y	= autoZoom_y;
			nodeChild->autoZoom.z	= autoZoom_z;

			nodeChild->scroll		= scroll;
			
			nodeChild->triggerOnHi.x	= triggerOnHi_x;
			nodeChild->triggerOnHi.y	= triggerOnHi_y;
			nodeChild->triggerOnHi.z	= triggerOnHi_z;

			nodeChild->triggerOnLow.x	= triggerOnLow_x;
			nodeChild->triggerOnLow.y	= triggerOnLow_y;
			nodeChild->triggerOnLow.z	= triggerOnLow_z;
			
			nodeChild->triggerLevelHi.x = triggerLevelHi_x;
			nodeChild->triggerLevelHi.y = triggerLevelHi_y;
			nodeChild->triggerLevelHi.z = triggerLevelHi_z;

			nodeChild->triggerLevelLow.x = triggerLevelLow_x;
			nodeChild->triggerLevelLow.y = triggerLevelLow_y;
			nodeChild->triggerLevelLow.z = triggerLevelLow_z;

			//node->triggerCommand = triggerCommand;		//zz debug
			nodeChild->recordID = (int)recordID;
			
			nodeChild->proximity.x = (int)proximity_x;
			nodeChild->proximity.y = (int)proximity_y;
			nodeChild->proximity.z = (int)proximity_z;
			
			nodeChild->proximityRadius.x = proximityRadius_x;
			nodeChild->proximityRadius.y = proximityRadius_y;
			nodeChild->proximityRadius.z = proximityRadius_z;

			nodeChild->drawCurrentValue.x = drawCurrentValue_x;
			nodeChild->drawCurrentValue.y = drawCurrentValue_y;
			nodeChild->drawCurrentValue.z = drawCurrentValue_z;

			nodeChild->selectionSet.x = selectionSet_x;
			nodeChild->selectionSet.y = selectionSet_y;
			nodeChild->selectionSet.z = selectionSet_z;

			nodeChild->size = lineSize;
		}

	if (nodeChild != NULL)
		printf("id: %d   color: %d   recordID: %d\n", nodeChild->id, nodeChild->colorIndex, nodeChild->recordID);	//debug zz

	}

	data->map.nodeRootIndex = data->map.nodeRootCount - 1;
	data->map.selectedIndex = data->map.nodeRootIndex;
	data->map.selectedNode = data->map.node[data->map.selectedIndex];
	data->map.activeNode = data->map.selectedNode;
}

//MB-End//



//-----------------------------------------------------------------------------
void npLoadChannelFile (char* buffer, int wordSize, int size, void* dataRef)
{
	int i = 0, j = 0;
	int curs = 0;
	int localCycleIndex = 0;
	int localChannelIndex = 0;
	int count = 0;
	int scanNumRet = 0;
	int numExpected = 30;

	int cycleCount = 0;
	char chA[64],chB[64],chC[64],chD[64],chE[64],chF[64],chG[64],chH[64],chI[64],chJ[64];
	char chBA[64],chBB[64],chBC[64],chBD[64],chBE[64],chBF[64],chBG[64],chBH[64],chBI[64],chBJ[64];
	char chCA[64],chCB[64],chCC[64],chCD[64],chCE[64],chCF[64],chCG[64],chCH[64],chCI[64],chCJ[64];
	
	float ch[30];

	NPnodePtr node = NULL;

	pData data = (pData) dataRef;

	for (i=0; i < 30; i++)
		ch[i] = 0;

	// parse string and copy contents into the channels buffer
	// update channelSize and channelIndex,

	// iterate-loop through buffer

	// if (newline) then increment localCycleIndex++
	// and reset localChannelIndex = 10
	// process line of data, repeat

	// convert from string to float
	// search for first line return

	//*** create a dummy procedure to fill with random data...
	
	while (scanNumRet != -1)
	{

		curs = npSeekToNextLine( buffer );

		buffer = &buffer[curs];				// add err checking for end of file, debug, zz

//		printf( "First Real char: %c\n", buffer[0] );

		scanNumRet = sscanf ( buffer, "%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^','],%[^',']",
							&chA,&chB,&chC,&chD,&chE,&chF,&chG,&chH,&chI,&chJ,
							&chBA,&chBB,&chBC,&chBD,&chBE,&chBF,&chBG,&chBH,&chBI,&chBJ,
							&chCA,&chCB,&chCC,&chCD,&chCE,&chCF,&chCG,&chCH,&chCI,&chCJ);
		
		if( scanNumRet != numExpected )
		{
			if( scanNumRet == -1 )
			{
				printf( "End of File\n" );
			// update channelSize and channelIndex,
			}
			else
				printf( "WARNING: Bad Format\n" );
		}
		else
		{
			ch[0] = (float) atof(chA);
			ch[1] = (float) atof(chB);
			ch[2] = (float) atof(chC);
			ch[3] = (float) atof(chD);
			ch[4] = (float) atof(chE);
			ch[5] = (float) atof(chF);
			ch[6] = (float) atof(chG);
			ch[7] = (float) atof(chH);
			ch[8] = (float) atof(chI);
			ch[9] = (float) atof(chJ);

			ch[10] = (float) atof(chA);
			ch[11] = (float) atof(chB);
			ch[12] = (float) atof(chC);
			ch[13] = (float) atof(chD);
			ch[14] = (float) atof(chE);
			ch[15] = (float) atof(chF);
			ch[16] = (float) atof(chG);
			ch[17] = (float) atof(chH);
			ch[18] = (float) atof(chI);
			ch[19] = (float) atof(chJ);

			ch[20] = (float) atof(chA);
			ch[21] = (float) atof(chB);
			ch[22] = (float) atof(chC);
			ch[23] = (float) atof(chD);
			ch[24] = (float) atof(chE);
			ch[25] = (float) atof(chF);
			ch[26] = (float) atof(chG);
			ch[27] = (float) atof(chH);
			ch[28] = (float) atof(chI);
			ch[29] = (float) atof(chJ);
	
			data->io.channel[10][localCycleIndex] = ch[0];
			data->io.channel[11][localCycleIndex] = ch[1];
			data->io.channel[12][localCycleIndex] = ch[2];
			data->io.channel[13][localCycleIndex] = ch[3];
			data->io.channel[14][localCycleIndex] = ch[4];
			data->io.channel[15][localCycleIndex] = ch[5];
			data->io.channel[16][localCycleIndex] = ch[6];
			data->io.channel[17][localCycleIndex] = ch[7];
			data->io.channel[18][localCycleIndex] = ch[8];
			data->io.channel[19][localCycleIndex] = ch[9];

			data->io.channel[20][localCycleIndex] = ch[10];
			data->io.channel[21][localCycleIndex] = ch[11];
			data->io.channel[22][localCycleIndex] = ch[12];
			data->io.channel[23][localCycleIndex] = ch[13];
			data->io.channel[24][localCycleIndex] = ch[14];
			data->io.channel[25][localCycleIndex] = ch[15];
			data->io.channel[26][localCycleIndex] = ch[16];
			data->io.channel[27][localCycleIndex] = ch[17];
			data->io.channel[28][localCycleIndex] = ch[18];
			data->io.channel[29][localCycleIndex] = ch[19];

			data->io.channel[30][localCycleIndex] = ch[10];
			data->io.channel[31][localCycleIndex] = ch[11];
			data->io.channel[32][localCycleIndex] = ch[12];
			data->io.channel[33][localCycleIndex] = ch[13];
			data->io.channel[34][localCycleIndex] = ch[14];
			data->io.channel[35][localCycleIndex] = ch[15];
			data->io.channel[36][localCycleIndex] = ch[16];
			data->io.channel[37][localCycleIndex] = ch[17];
			data->io.channel[38][localCycleIndex] = ch[18];
			data->io.channel[39][localCycleIndex] = ch[19];

			printf( "%f %f %f %f %f %f %f %f %f %f\n", 
				ch[0], ch[1], ch[2], ch[3], ch[4], ch[5], ch[6], ch[7], ch[8], ch[9] );

			localCycleIndex++;
		}
	}

	// update the channel array parameters
	data->io.channelSize = localCycleIndex - 8;							// use cyclecount to verify debug, zz
	if (data->io.channelSize < 1)
		data->io.channelSize = 1;
 
	data->io.channelIndex = 0;
	data->io.channelOldest = 0;
}


//-----------------------------------------------------------------------------
void npSetChannelFile (char* fileName, int wordSize, int size, void* dataRef)
{
	FILE* file = NULL;
	size_t count = 0;
	int i = 0;
	int byteCount = 0;
	char* bufferIndex = 0;

	
	NPnodePtr node = NULL;
	pData data = (pData) dataRef;

	char* str = (char*)malloc(260);
	char* path = (char*)malloc(260);
	char* buffer = (char*)malloc(kNPmapFileWriteBufferSize);

	if (buffer == NULL)
	{
		printf ("4214 error malloc failed cannot write file");
		npFileClose(file, dataRef);
		free (str);
		free (path);
		free (buffer);
		return;
	}

	strcpy (str, fileName);

	// get the Current Working Directory
	npGetCWD (path, &size);

	printf("CWD path : ");
	for(i = 0; i < size; i++)
		printf ("%c", path[i]);
	printf("\n");

	printf("\nRequired Length : %d\n", (strlen(path) + strlen(fileName))); 
	str = strcat(path, fileName);


	// open/create and write to the file, "w+" overwrites existing files
	file = npFileOpen(str, "r", dataRef);

	if( file != NULL )   // check to see if file exists		 debug, ll
	{
		printf("\nReading File\n"); // was "File is not NULL"			debug, zz
		npFileRewind(file);
		count = npFileRead(buffer, 1, kNPmapFileWriteBufferSize, file, dataRef);
		printf("\nSuccessfully Read %d bytes \n", count);

		//print the File Contents
		printf("File Contents : \n");
		size = count;
		if ( size > 2048) size = 2048;
			for( i = 0; i < size; i++ )
				printf("%c", buffer[i]);

		if (size < 1)
			printf("\n Error : Loading State : File size is zero \n");
		else
		{
			printf("\n Loading State \n");
			npLoadChannelFile (buffer, 1, count, dataRef);
		}

		npFileClose(file, dataRef);
	}
	else
		printf("\n File cannot open file \n");



	free (str);
	free (path);
	free (buffer);
}

//-----------------------------------------------------------------------------
void npSetMapFile (char* fileName, int wordSize, int size, void* dataRef)
{
	FILE* file = NULL;
	size_t count = 0;
	int i = 0;
	int byteCount = 0;
	char* bufferIndex = 0;

	
	NPnodePtr node = NULL;
	pData data = (pData) dataRef;

	char* str = (char*)malloc(260);
	char* path = (char*)malloc(260);
	char* buffer = (char*)malloc(kNPmapFileWriteBufferSize);

	if (buffer == NULL)
	{
		printf ("4214 error malloc failed cannot write file");
		npFileClose(file, dataRef);
		free (str);
		free (path);
		free (buffer);
		return;
	}

	strcpy (str, fileName);

	// get the Current Working Directory
	npGetCWD (path, &size);

	printf("CWD path : ");
	for(i = 0; i < size; i++)
		printf ("%c", path[i]);
	printf("\n");

	printf("\nRequired Length : %d\n", (strlen(path) + strlen(fileName))); 
	str = strcat(path, fileName);


	// open/create and write to the file, "w+" overwrites existing files
	file = npFileOpen(str, "r", dataRef);

	if( file != NULL )   // check to see if file exists		 debug, ll
	{
		printf("\nReading File\n"); // was "File is not NULL"			debug, zz
		npFileRewind(file);
		count = npFileRead(buffer, 1, kNPmapFileWriteBufferSize, file, dataRef);
		printf("\nSuccessfully Read %d bytes \n", count);

		//print the File Contents
		printf("File Contents : \n");
		size = count;
		if ( size > 2048) size = 2048;
		for( i = 0; i < size; i++ )
			printf("%c", buffer[i]);

		if (size < 1)
			printf("\n Error : Loading State : File size is zero \n");
		else
		{
			printf("\n Loading State \n");
			npLoadMapFile (buffer, 1, count, dataRef);
		}
		
		npFileClose(file, dataRef);
	}
	else
		printf("\n File cannot open file \n");


	free (str);
	free (path);
	free (buffer);
	return;
}


//-----------------------------------------------------------------------------
void npTestMapFile (char* fileName, int wordSize, int size, void* dataRef)
{
	FILE* file = NULL;
	size_t count = 0;
	int i = 0;
	int byteCount = 0;
	char* bufferIndex = NULL;
	char* charPtrA = NULL;
	char* charPtrB = NULL;

	NPnodePtr node = NULL;
	pData data = (pData) dataRef;

	char* str = (char*)malloc(260);
	char* buffer = (char*)malloc(kNPmapFileWriteBufferSize);

	if (buffer == NULL || str == NULL)
	{
		printf ("4214 error malloc failed cannot write file");
		free (str);
		return;
	}

	strcpy (str, fileName);  // this is for combining a file path with filename, zz

	// open or create and write to the file, "w" overwrites existing files
	file = npFileOpen(str, "w+", dataRef);

	if( file != NULL )						//check to see if file exists
		printf("\nFile Opened\n");
	else
	{
		printf("\nFile cannot be created\n");							//debug, add err reporting, zz
		return;
	}
	
	bufferIndex = buffer + strlen(buffer);
	bufferIndex += sprintf(bufferIndex, "\n");		//begin file with new line

	// fill the buffer with the map of current state
	count = npGetMapFile (bufferIndex, count, node, dataRef);

	// write the nodes to file
	printf("\nSaving State\n");
	count += npFileWrite (buffer, 1, strlen(buffer), file, dataRef);		//debug, strlen should be count instead, zz

	// write pin data
	charPtrA = ( buffer + ((kNPmapFileWriteBufferSize * 10) / 16));		//debug, bizarre way of merging multiple arrays together, zz
	byteCount = (int)(data->io.write[kNPpin] - charPtrA);
	count += npFileWrite (charPtrA, 1, byteCount, file, dataRef);

	// write child id lists
	charPtrA = ( buffer + ((kNPmapFileWriteBufferSize * 12) / 16));
	byteCount = (int)(data->io.write[kNPchild] - charPtrA);
	count += npFileWrite (charPtrA, 1, byteCount, file, dataRef);

	// write camera data
/*	charPtrA = ( buffer + ((kNPmapFileWriteBufferSize * 8) / 16));
	byteCount = (int)(data->io.write[kNPcamera] - charPtrA);
	count += npFileWrite (charPtrA, 1, 100, file, dataRef);
*/

	// print out the byte count and first few bytes of the file
	if (count)	//== strlen(buffer) )
	{
		printf("\nSuccessfully Wrote %d bytes to file\n", count);
		
		// read the file
		npFileRewind(file);
		size = npFileRead (buffer, 1, 256, file, dataRef);

		//Print File Contents
		printf("File Contents : \n");
		for( i = 0; i < size; i++ )
			printf("%c", buffer[i]);
	}
	else
	{
		printf("\nFile Write Failure\n");
		npFileClose(file, dataRef);
		return; // Not working, return -1
	}


	npFileClose(file, dataRef);
	free (str);
	free (buffer);
	return;
}


//-----------------------------------------------------------------------------
// if not saved then opens SaveAs dialog before reseting app
void npCloseFileMapFile (int size, void* dataRef)
{
	return;
}


//-----------------------------------------------------------------------------
int npGetMapFileAttrib (int attrib, void* dataRef)
{
	switch (attrib)
	{
	case kNPversion: ; break;
	case kNPtableCount: ; break;	// number of struct equivalent data tables 
	case kNPrecordCount: ; break;	// total number of records
	case kNPnodeRootCount: ; break;
	case kNPgraphCount: ; break;	// number of graph elements connect nodes
	case kNPsize: ; break;			// RAM used by map file

	default : break;
	}

	return 0;
}

//-----------------------------------------------------------------------------
void npSetMapFileAttrib (int attrib, void* dataRef)
{
	return;
}




//-----------------------------------------------------------------------------
// keep track of a bufferIndex for each list
// write list of children
// write node or node list
// write node specific data to a separate list, ie pin, camera, etc
// write list of channels,						channels NOT YET IMPLEMENTED, debug, zz
//-----------------------------------------------------------------------------
int npWriteNode (char* buffer, void* nodeRef, void* dataRef)
{
	int i			= 0;
	int n			= 0;
	int total		= 0;
	int parentID	= 0;

	pData data = (pData) dataRef;

	// pointers to the node data, update to be a generic list, debug, zz
	char* nodePtr		= data->io.write[kNPnode];		//	same as buffer
	char* cameraPtr		= data->io.write[kNPcamera];	//	kNPmapFileWriteBufferSize * 8 / 16;
	char* gridPtr		= data->io.write[kNPgrid];		//	kNPmapFileWriteBufferSize * 9 / 16;
	char* pinPtr		= data->io.write[kNPpin];		//	kNPmapFileWriteBufferSize * 10 / 16;
	char* childPtr		= data->io.write[kNPchild];		//	kNPmapFileWriteBufferSize * 12 / 16;
	char* channelPtr	= data->io.write[kNPchannel];	//	kNPmapFileWriteBufferSize * 14 / 16;


	NPnodePtr node = (NPnodePtr) nodeRef;
	NPnodePtr parent = NULL;
	NPnodePtr child = NULL;

	NPcameraPtr camera = NULL;
	NPgridPtr grid = NULL;
	NPpinPtr pin = NULL;

	
	if(node->parent != NULL)
	{
		parent = node->parent;
		parentID = parent->id;
	}
	else
		parent = node;

	if (0)//buffer != nodePtr)												// debug, zz
		printf ("\n writePtr %p %p %p %p %p %p \n", 
				nodePtr, cameraPtr, gridPtr, pinPtr, childPtr, channelPtr);


	// create a list of child->id, the list id is the same as this node->id
	// used to store memory pointers as an id list for file or db exchange
	if (node->childCount)
	{
		n += sprintf ((childPtr + n), "%d,", node->id );

		for (i=0; i < node->childCount; i++)
		{
			child = node->child[i];

			if (child != NULL)
			{
				// if first item, no comma in front, otherwise insert a comma
				if (i == 0)
					n += sprintf ((childPtr + n), "%d", child->id );
				else
					n += sprintf ((childPtr + n), ",%d", child->id );
			}
		}
	
		n += sprintf ((childPtr + n), "\n", node->id );

		data->io.write[kNPchild] += n;						//					debug zz
		total += n;
		n = 0;
	}



	n += sprintf ((nodePtr + n), "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
	node->id,					//id is unique to this node
	node->type,		
	node->id,					// node->data->id uses this node id
	
	node->selected,
	parentID,					// parent id replaces pointer to parent
	node->branchLevel,
	node->id,					// node->id is same as childList->id
	node->childIndex,
	node->childCount,
	
	node->channel.x,
	node->channel.y,	
	node->channel.z,
	
	node->channelIndex.x,
	node->channelIndex.y,
	node->channelIndex.z,

	node->averageType.x,
	node->averageType.y,
	node->averageType.z,
	
	node->sampleInterval.x,
	node->sampleInterval.y,
	node->sampleInterval.z );


	n += sprintf ((nodePtr + n), ",%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f", 
	node->rotate.angle, 
	node->rotate.x,
	node->rotate.y,
	node->rotate.z,
	
	node->scale.x,
	node->scale.y,
	node->scale.z,
	
	node->translate.x,
	node->translate.y,
	node->translate.z,

	node->nodeOrigin.x,
	node->nodeOrigin.y,
	node->nodeOrigin.z,
	
	node->rotateRate.x,
	node->rotateRate.y,
	node->rotateRate.z,

	node->rotateRad.x,
	node->rotateRad.y,
	node->rotateRad.z );

	n += sprintf ((nodePtr + n), ",%f,%f,%f,%f,%f,%f,%f,%f,%f", 
	node->scaleRate.x,
	node->scaleRate.y,
	node->scaleRate.z,

	node->translateRate.x,
	node->translateRate.y,
	node->translateRate.z,

	node->translateVect.x,
	node->translateVect.y,
	node->translateVect.z
	 );

	n += sprintf ((nodePtr + n), ",%d,%d,%f,%f,%f,%d,%d,%d,%d,%d,%d",
	node->surfaceType,
	node->shaderType,

	node->lineWidth,
	node->pointSize,
	node->circle,

	node->colorIndex,
	
	node->color.r,
	node->color.g,
	node->color.b,
	node->color.a,

	node->colorFade );

	n+= sprintf ((nodePtr + n), ",%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
	node->hide,
	node->freeze,
	node->center,
	node->centerPerGrid,

	node->autoZoom.x,
	node->autoZoom.y,
	node->autoZoom.z,

	node->scroll,
	
	node->triggerOnHi.x,
	node->triggerOnHi.y,
	node->triggerOnHi.z,

	node->triggerOnLow.x,
	node->triggerOnLow.y,
	node->triggerOnLow.z );

	n += sprintf ((nodePtr + n), ",%f,%f,%f,%f,%f,%f,%d,%f,%f,%f,%f,%f,%f",
	node->triggerLevelHi.x,
	node->triggerLevelHi.y,
	node->triggerLevelHi.z,

	node->triggerLevelLow.x,
	node->triggerLevelLow.y,
	node->triggerLevelLow.z,

	//node->triggerCommand,			//zz debug
	node->recordID,
	
	node->proximity.x,
	node->proximity.y,
	node->proximity.z,
	
	node->proximityRadius.x,
	node->proximityRadius.y,
	node->proximityRadius.z );

	n += sprintf ((nodePtr + n), ",%d,%d,%d,%d,%d,%d,%d\n",
	node->drawCurrentValue.x,
	node->drawCurrentValue.y,
	node->drawCurrentValue.z,
	node->selectionSet.x,
	node->selectionSet.y,
	node->selectionSet.z,

	node->size
	);	

	data->io.write[kNPnode] += n;						//					debug zz
	total += n;
	n = 0;

	if (node->type == kNodePin)
	{
		pin = (NPpinPtr) node->data;

		n += sprintf ((pinPtr + n), "%d,%d,%d,%f,%f,%d,%d,%d,%d,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%d\n",
			pin->id,					//id is unique to this node
			pin->type,		
			pin->textureID,					// node->data->id uses this node id
			pin->innerRadius,
			pin->outerRadius,
			pin->radiusRatioIndex,
			pin->screenSizeIndex,
			pin->slices,
			pin->stacks,
			pin->translateTex.x,
			pin->translateTex.y,
			pin->translateTex.z,
			pin->scaleTex.x,
			pin->scaleTex.y,
			pin->scaleTex.z,
			pin->rotateTex.x,
			pin->rotateTex.y,
			pin->rotateTex.z,
			pin->rotateTex.angle,
			pin->size );
	}

	data->io.write[kNPpin] += n;						//					debug zz
	total += n;
	n = 0;

	if (node->type == kNodeCamera)
	{
		camera = (NPcameraPtr) node->data;

		n += sprintf ((cameraPtr + n), "%d,%d,%d,%d,%d,%d,%f,%f,%d,%d,%d,%f,%f,%f,%f,%f,%d,%d\n",
			camera->id,					//id is unique to this node
			camera->type,

			camera->textureID,					// node->data->id uses this node id

			camera->format,
			camera->interlaced,
			camera->stereo,
			camera->aspectRatio,
			camera->fps,
			camera->colorSpace,
			camera->width,
			camera->height,

			camera->fov,
			camera->clipNear,
			camera->clipFar,
			camera->aperture,
			camera->exposure,
			camera->sensorType,

			camera->size );
	}

	data->io.write[kNPcamera] += n;						//					debug zz
	total += n;
	n = 0;

	return total;
}

